/***********************************************************************
 *                                                                     *
 *                              Bookmarks                              *
 *                                                                     *
 ***********************************************************************
  Bookmarks are multi-level. There are 4 kinds of bookmark entries:
  - bookmark folder  'F'
  - bookmark         'B'
  - folder end       '>'

  The general on-disk format is:

    [size] [type] [page] [title]

  If type is '<' or '>', then the page and title fields are missing.
  For 'F' type entries, the page field is missing. All fields are
  present for 'B' entries.

  The whole bookmark section is a single entity. There is no table etc.
  associated with it.

  The input format for bookmark files is:

     BookmarkTitle: Title Here
     BookmarkLevel: 1
     BookmarkPageNumber: PageNumber or 0 for bookmark-folders.

  This is the output format of the pdftk program.
 ***********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "bmk.h"


static bmk_t *bmks;

static void strrem(char *str,int start, int end)
{
  memmove(str+start, str+end, strlen(str+end)+1);
}

static void strip_spc(char *str)
{
  unsigned char *K;
  int i,L;
  K= str;
  for(i=0;K[i];i++)
    if (K[i]!=127 && K[i]>' ') break;
  if (i!=0) strrem(K, 0, i);
  L= strlen(K);
  while(L>0 && (K[L-1]==127 || K[L-1]<=' ')) L--;
  K[L]= 0;
}

static void strip_dots(char *K)
{
  int L;
  L= strlen(K);
  while(L>0 && K[L-1]=='.') L--;
  K[L]= 0;
}

static void add_bmk(char *line)
{
   bmk_t *B;
   strip_spc(line);
   strip_dots(line);
   B= calloc(1, sizeof(*B));
   B->title= strdup(line);
   B->prev= bmks->prev;
   B->next= bmks;
   bmks->prev->next= B;
   bmks->prev= B;
}

static int strpfx(char* str, const char *pfx)
{
  int Ls, Lp;
  int i;
  Ls= strlen(str); Lp= strlen(pfx);
  if (Ls<Lp) return 0;
  for(i=0;i<Lp;i++) if (str[i]!=pfx[i]) return 0;
  return 1;
}

static void bmk_line(char *line)
{
  int i;
  static const char *fields[]= 
  {
    "BookmarkTitle:", "BookmarkLevel:", "BookmarkPageNumber:", NULL
  };
  for(i=0;fields[i];i++)
   if (strpfx(line, fields[i]))
   {
     strrem(line,0,strlen(fields[i]));
     switch(i)
     {
     case 0: add_bmk(line); return;
     case 1: if (bmks->prev!=bmks) bmks->prev->level= atoi(line); return;
     case 2: if (bmks->prev!=bmks) bmks->prev->page= atoi(line); return;
     }
   }
}


static void parse_data(const char *fn)
{
  FILE *inp;
  char line[2000];

  inp= fopen(fn,"r");
  if (!inp)
   exit(fprintf(stderr,"can't open bookmark file %s: %s\n", fn,
                       strerror(errno)));

  while(fgets(line, sizeof(line), inp))
    if (strpfx(line,"Bookmark")) bmk_line(line);
  fclose(inp);
}

void dbg_print_bmklist(bmk_t *list,int L)
{
  bmk_t *B;
  int i;
  for(B=list->next;B!=list;B=B->next)
  {
    for(i=0;i<L;i++) printf("  ");
    printf("%c %s [%d,%d]\n", B->type ? B->type: 'N', B->title,B->page,B->level);
    if (B->children) dbg_print_bmklist(B->children, L+1);
  }
}

void dbg_print_bmks()
{
  bmk_t *B;
  dbg_print_bmklist(bmks,1);
}


/* This function collects child bookmarks into their parents. A bookmark
   is the child of another if they are sequential and their levels are
   also sequential. */

static void recover_struct(bmk_t *list)
{
   bmk_t *P,*B,*E,*C,*L;
   for(P=list->next;P!=list;P=P->next)
   {
      B= P->next;
      if (B==list || B->level<=P->level) continue ;

      for(E=B->next;E!=list;E=E->next)
        if (E->level<=P->level) break;
      L= E->prev;

      C= P->children= calloc(1,sizeof(*P->children));
      
      /* remove [B .. L] from the main list */
      B->prev->next= L->next;
      L->next->prev= B->prev; 

      /* put them in the child list */
      C->next= B; C->prev= L;
      B->prev= C; L->next= C;

      recover_struct(C);
   }
}

static int imax(int a,int b) { return a>b ? a: b; }

static void expunge_empty_children(bmk_t *B)
{
   if (B->children && B->children->next==B->children)
   {
      free(B->children);
      B->children= NULL;
   }
}

static void merge_bmks(bmk_t *B, bmk_t *D)
{
   bmk_t *C, *F, *L;

   B->page= imax(B->page, D->page);

   expunge_empty_children(B);
   expunge_empty_children(D);

   if (!B->children) { B->children= D->children; goto done; }
   if (!D->children) goto done;

   F= D->children->next;
   L= D->children->prev; 
   C= B->children;

   F->prev= C->prev;
   L->next= C;

   C->prev->next= F;
   C->prev= L;

   free(D->children);
done:
   free(D->title);
   free(D);
}

/* This function does the main work. If it finds two bookmarks with the
   same title, it unifies them. */

void unify_bmks(bmk_t *list)
{
   bmk_t *B, *D,*N;
   for(B=list->next;B!=list;B=B->next)
     for(D=B->next;D!=list;D=N)
     {
       N= D->next;
       if (strcmp(D->title,B->title)) continue;
       D->prev->next= D->next;
       D->next->prev= D->prev;
       merge_bmks(B, D);
     }

   for(B=list->next;B!=list;B=B->next)
     if (B->children && B->children->next!=B->children) 
        unify_bmks(B->children);
} 

static void assign_types(bmk_t *list)
{
  bmk_t *B;
  for(B=list->next;B!=list;B=B->next)
  {
    if (B->children) B->type= 'F';
    else if (B->page) B->type= 'B';
    if (B->children) assign_types(B->children);
  }
}


bmk_t* bmk_parse(const char *fn)
{
  bmk_t *top;

  bmks= calloc(1,sizeof(*bmks));
  bmks->prev= bmks->next= bmks;
  parse_data(fn);
  recover_struct(bmks);
  unify_bmks(bmks);
  assign_types(bmks);

  top= calloc(1,sizeof(*top));
  top->children= bmks;
  
  return top;
}
